/* 
 * Linkoping Intelligent Communication of Knowledge System (LINCKS)
 *      Copyright (C) 1993, Lin Padgham, Ralph Rnnquist
 *       Department of Computer and Information Sciences
 *		University of Linkoping, Sweden
 *		    581 83 Linkoping, Sweden
 *		       lincks@ida.liu.se
 *
 * These collective LINCKS programs are free software; you can 
 * redistribute them and/or modify them under the terms of the GNU
 * General Public License as published by the Free Software Foundation,
 * version 2 of the License.
 *
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the programs; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * MODULE NAME:		aimexport.c
 *
 * SCCSINFO:		@(#)aimexport.c	1.3 12/20/93
 *
 * ORIGINAL AUTHOR:	David Partain, 1990-06-06
 *
 * MODIFICATIONS:
 *	1993-11-26 Martin Sjlin and David Partain.  Reworked several
 *			things, but the gist of it is that there are now
 *			two additional configuration options.  (1) you
 *			set whether to escape characters for LaTeX.  (2)
 *			you say whether you want placeholders included.
 *      <list mods here with name and date>
 *
 * DESCRIPTION:
 *	This module will contain the functions that will enable
 *	the user to export a window as text (either LaTeX or straight)
 *	text.
 */

/*********************************************************************
 * INCLUDES:
 *********************************************************************/
#include <stdio.h>
#include <ctype.h>
#ifndef sequent
#include <malloc.h>
#endif	/* sequent */
#include <string.h>
#if (defined(ultrix)) || (defined(sequent))
#include <sys/types.h>	/* stat.h needs it */
#endif
#include <sys/stat.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Form.h>

#include "aimtypes.h"
#include "xconfig.h"		/* for DBDIR */
#include "systypes.h"

/*********************************************************************
 * EXTERNALLY-CALLABLE ROUTINES FOUND IN THIS MODULE:
 *********************************************************************/
Widget build_export( /* reference_structure *rsp */ );

/*********************************************************************
 * EXTERNALLY-AVAILABLE DATA FOUND IN THIS MODULE:
 *********************************************************************/
/* none */

/*********************************************************************
 * EXTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
/* xstuff.c */
extern XtArgVal CVfont( /* char *vp */ );
extern void popup_message( /* varargs */ );

/* xdumpref.c */
extern void whack_widl( /* CALLBACK PROTOCOL */ );

#ifdef __ultrix
extern char *strdup( /* char *incoming */ );
#endif

#ifdef sequent
extern char *malloc();
extern void free();
#endif  /* sequent */

/*********************************************************************
 * EXTERNAL DATA STRUCTURES USED BY THIS MODULE:
 *********************************************************************/
extern Widget toplevel;		/* xstuff.c */

/*********************************************************************
 * LOCAL DEFINES, STRUCTS, TYPEDEFS, ETC.:
 *********************************************************************/
#ifndef MAXPATHLEN		/* also in lincks.h */
#define MAXPATHLEN 512
#endif /* MAXPATHLEN */

/* file which holds font translations from their X to LaTeX forms */
#define FONT_TRANSLATION_FILE ".fonttrans"

/*
 * In escape_sp_chars, these two are used to signify a left quotation
 * mark and to signify a right quotation mark.
 */
#define LEFT 0
#define RIGHT 1

/* these indicate the depth of the LaTeX file to be created.  */
#define FIRST_LEVEL 1
#define SECOND_LEVEL 2
#define THIRD_LEVEL 3
#define FOURTH_LEVEL 4
#define FIFTH_LEVEL 5

/* print_part prints all of these to the file */
#define PART 1
#define PARAGRAPH 2
#define SECTION 3
#define SUBSECTION 4
#define SUBSUBSECTION 5
#define TRAILER 6
#define LETTERHEAD 7
#define LETTERTAIL 8
#define SQUIGGLES 9

/* when creating title, these are used to choose parameters.  */
#define TITLE 1
#define DATE 2
#define AUTHOR 3

#define BUF_SIZE 100		/* for strings */
#define MAX_STACK 10		/* number of LaTeX fonts on stack */
#ifndef MAX_WINS		/* they might have it on the command line */
#define MAX_WINS 25		/* number of LaTeX fonts on stack */
#endif /* MAX_WINS */

/*
 * This structure holds a pair of information.  When an X font is found,
 * its corresponding LaTeX font is loaded into "fontname", and the
 * number of {'s to close is put into "completion".
 */
typedef struct x_LaTeXfonts {
  char *fontname;
  int completion;
} LaTeXfonts;

typedef struct __textdump {
  enum DocStyle {
    article = 0, letter = 1
  } docstyle;
  enum Structure {
    deep, flat
  } structure;
  enum FontSize {
    small = 10, medium = 11, large = 12
  } fontsize;
  char *outfile;
  enum Placeholders {
    yes, no
  } placeholders;
} _textdump;

typedef struct _open_win_struct {
  reference_structure *ref;
  _textdump *textdump;
  Widget dumptypew;
  Widget stylew;
  Widget fontw;
  Widget structw;
  Widget escapew;
  Widget placeholderw;
  Widget shellw;
  Widget statusw;
  Widget outputw;
} open_win_struct;

typedef struct _filefontinfo {
  struct _filefontinfo *next;
  char *Xfontname;
  char *LaTeXname;
  int squigglies;
} filefontinfo;

/*********************************************************************
 * INTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
static Widget setup_dump( /* reference_structure *ref, int index */ );
static void num_levels( /* infonode *inodeptr, int level */ );
static void get_title( /* FILE *outfile, identry *idptr, int index */ );
static int check_identry( /* identry *idp, char **title, char **author,
			     char **date, int index */ );
static int check_infonode( /* infonode *inp, char **title, char **author,
			      char **date, int index */ );
static char *substitute_title( /* identry *idptr */ );
static int compare_strings( /* char *string, char *to_find, int *mark */ );
static void LaTeX_node( /* int index,infonode *infoptr, FILE *outfile */ );
static void print_node( /* int index, infonode *info, FILE *outfile */ );
static char *attval_plus( /* infonode *infoptr,char *localfont */ );
static int my_level( /* identry *identptr */ );
static char *escape_sp_chars( /* char *text */ );
static void find_sect_title( /* int index, infonode *infoptr, 
				int check_level, FILE *outfile */ );
static void clear_printed( /* infonode *inodeptr */ );
static int add_formats( /* infonode *infoptr */ );
static void delete_formats( /* int where */ );
static int load_file_fonts();
static filefontinfo *new_filefontinfo();
static filefontinfo *get_format( /* char *lookfor */ );
static _textdump *initialize();
static int is_open( /* reference_structure *ref, _textdump *textdump */ );
static int set_outfile( /* int index */ );
static char *get_file_name();
static void get_header( /* int index, FILE *outfile */ );
static void make_art_title( /*  FILE *outfile, char *title,
				char *author, char *date */ );
static void print_part( /* int index, FILE *outfile, char *part, int which */ );

static void execute_dump( /* CALLBACK PROTOCOL */ );
static void toggle_dumptype( /* CALLBACK PROTOCOL */ );
static void toggle_style( /* CALLBACK PROTOCOL */ );
static void toggle_font( /* CALLBACK PROTOCOL */ );
static void toggle_structure( /* CALLBACK PROTOCOL */ );
static void toggle_escape( /* CALLBACK PROTOCOL */ );
static void toggle_phold( /* CALLBACK PROTOCOL */ );
static void free_stuff( /* CALLBACK PROTOCOL */ );
static void close_dump( /* CALLBACK PROTOCOL */ );

/*********************************************************************
 * INTERNAL (STATIC) DATA:
 *********************************************************************/
static int prevlevel = 0;	/* level of previous leaf node */
static int thislevel = 0;	/* level of this leaf node */
static int depth_signal = 0;	/* signals that depth > 4 */
static int indent = 0;		/* number spaces to indent == indent * 3 */
static LaTeXfonts fontstack[MAX_STACK];/* LaTeX fonts */
static int stkptr = -1;		/* top of stack pointer in LaTeXfonts */
static char *prevfont = NULL;	/* last LaTeX font written to file */
static int lefttoprint = 0;	/* number of open squigglies */

static enum EscType {
  doit, dont
} global_escape = doit;		/* whether to escape chars for LaTeX */

static enum DumpType {
  straight, LaTeX
} global_type = LaTeX;		/* whether to dump as straight text */

static int global_level = 0;	/* global for number of levels */

static filefontinfo *fontinfo = NULL;
static open_win_struct open_win[MAX_WINS];

/*  */
/*******************************************************************
 * Function: Widget build_export(reference_structure *rsp)
 *
 *  This function builds the LaTeX file.  Creates a hierarchical
 *  LaTeX file which can be edited by the user after creation.
 *
 *  Modifications:
 *	<list mods with name and date>
 */
Widget build_export(rsp)
  reference_structure *rsp;	/* ptr to ref structure being displayed */
{
  _textdump *textdump = NULL;
  int index;

  if (!rsp) {
    (void)fprintf(stderr,
		  "\nNull reference structure:  cannot create LaTeX file.\n");
    return NULL;
  }
  textdump = initialize();
  if ((index = is_open(rsp, textdump)) == -1) {
    return NULL;
  }
  return setup_dump(rsp, index);
}

/*  */
/*******************************************************************
 * Function: static Widget setup_dump(reference_structure *ref, int index)
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static Widget setup_dump(ref, index)
  reference_structure *ref;
  int index;
{
  Widget formw, titlew, stylelabelw, fontlabelw, 
	 structlabelw, escapelabelw, pholdlabelw,
  	 dumptypelabelw, doitw, closew, LaTeXlabelw;
  XtArgVal font;
  Dimension width;

  open_win[index].shellw = XtVaCreatePopupShell("Text Dump Shell",
						applicationShellWidgetClass,
						toplevel,
						XtNallowShellResize, True,
						NULL);

  formw = XtVaCreateManagedWidget("TextDumpForm", formWidgetClass,
				  open_win[index].shellw,
				  XtNdefaultDistance, (XtArgVal) 8,
				  NULL);

  titlew = XtVaCreateManagedWidget("TextDumpTitle", labelWidgetClass,
				   formw,
				   XtNlabel, "Text Dump Options",
				   XtNborderWidth, 0,
				   XtNjustify, XtJustifyCenter,
				   XtNresizable, True,
				   NULL);

  dumptypelabelw = XtVaCreateManagedWidget("TextDumpTypeLabel",
					   labelWidgetClass, formw,
					   XtNlabel, "Dump Type:",
					   XtNfromVert, titlew,
					   XtNborderWidth, 0,
					   NULL);

  open_win[index].dumptypew = XtVaCreateManagedWidget("TextDumpType",
						    commandWidgetClass, formw,
						      XtNwidth, 50,
					    XtNlabel, "nice long string here",
						      XtNfromVert, titlew,
						      NULL);

  LaTeXlabelw = XtVaCreateManagedWidget("TextDumpLaTeXlabel", labelWidgetClass,
					formw,
					XtNlabel, "LaTeX Document options:",
					XtNborderWidth, 0,
					XtNfromVert, dumptypelabelw,
					NULL);

  stylelabelw = XtVaCreateManagedWidget("TextDumpStyleLabel",
					labelWidgetClass,
					formw,
					XtNlabel, "Document style:",
					XtNborderWidth, 0,
					XtNfromVert, LaTeXlabelw,
					NULL);

  open_win[index].stylew = XtVaCreateManagedWidget("TextDumpStyle",
						   commandWidgetClass,
						   formw,
					    XtNlabel, "nice long string here",
						   XtNfromVert, LaTeXlabelw,
						   NULL);

  fontlabelw = XtVaCreateManagedWidget("TextDumpFontLabel", labelWidgetClass,
				       formw,
				       XtNlabel, "Font size:",
				       XtNborderWidth, 0,
				       XtNfromVert, stylelabelw,
				       NULL);

  open_win[index].fontw = XtVaCreateManagedWidget("TextDumpFont",
						  commandWidgetClass,
						  formw,
					    XtNlabel, "nice long string here",
						  XtNfromVert, stylelabelw,
						  NULL);

  structlabelw = XtVaCreateManagedWidget("TextDumpStructLabel",
					 labelWidgetClass,
					 formw,
					 XtNlabel, "Structure:",
					 XtNborderWidth, 0,
					 XtNfromVert, open_win[index].fontw,
					 NULL);

  open_win[index].structw = XtVaCreateManagedWidget("TextDumpStruct",
						    commandWidgetClass,
						    formw,
					    XtNlabel, "nice long string here",
						    XtNfromVert, fontlabelw,
						    NULL);

  escapelabelw = XtVaCreateManagedWidget("TextDumpEscapeLabel",
					 labelWidgetClass,
					 formw,
					 XtNlabel, "Escape chars?",
					 XtNborderWidth, 0,
					 XtNfromVert, open_win[index].structw,
					 NULL);

  open_win[index].escapew = XtVaCreateManagedWidget("TextDumpEscape",
						    commandWidgetClass,
						    formw,
					    XtNlabel, "nice long string here",
						    XtNfromVert, structlabelw,
						    NULL);

  pholdlabelw = XtVaCreateManagedWidget("TextDumpPHoldLabel",
					 labelWidgetClass,
					 formw,
					 XtNlabel, "Placeholders?",
					 XtNborderWidth, 0,
					 XtNfromVert, open_win[index].escapew,
					 NULL);

  open_win[index].placeholderw = XtVaCreateManagedWidget("TextDumpPHolder",
						    commandWidgetClass,
						    formw,
					    XtNlabel, "nice long string here",
						    XtNfromVert, escapelabelw,
						    NULL);

  open_win[index].outputw = XtVaCreateManagedWidget("TextDumpOutput",
						    dialogWidgetClass,
						    formw,
						    XtNlabel, "Output file:",
						    XtNvalue, "",
						    XtNfromVert, pholdlabelw,
						    NULL);

  open_win[index].statusw = XtVaCreateManagedWidget("TextDumpStatus",
						    labelWidgetClass,
						    formw,
				  XtNlabel, "Status: No operations initiated",
						    XtNborderWidth, 0,
						  XtNjustify, XtJustifyCenter,
					 XtNfromVert, open_win[index].outputw,
						    XtNresizable, True,
						    NULL);

  doitw = XtVaCreateManagedWidget("TextDumpDoit", commandWidgetClass,
				  formw,
				  XtNlabel, "Execute",
				  XtNfromVert, open_win[index].statusw,
				  NULL);

  closew = XtVaCreateManagedWidget("TextDumpClose", commandWidgetClass,
				   formw,
				   XtNlabel, "Close Window",
				   XtNfromVert, open_win[index].statusw,
				   XtNfromHoriz, doitw,
				   NULL);

  XtVaSetValues(open_win[index].dumptypew, XtNfromHoriz, stylelabelw, NULL);
  XtVaSetValues(open_win[index].stylew, XtNfromHoriz, stylelabelw, NULL);
  XtVaSetValues(open_win[index].fontw, XtNfromHoriz, stylelabelw, NULL);
  XtVaSetValues(open_win[index].structw, XtNfromHoriz, stylelabelw, NULL);
  XtVaSetValues(open_win[index].escapew, XtNfromHoriz, stylelabelw, NULL);
  XtVaSetValues(open_win[index].placeholderw, XtNfromHoriz, stylelabelw, NULL);

  if ((font = CVfont("9x15")) != (XtArgVal) 0) {
    XtVaSetValues(titlew, XtNfont, font, NULL);
    XtVaSetValues(doitw, XtNfont, font, NULL);
    XtVaSetValues(closew, XtNfont, font, NULL);
  }
  if ((font = CVfont("8x13")) != (XtArgVal) 0) {
    XtVaSetValues(stylelabelw, XtNfont, font, NULL);
    XtVaSetValues(fontlabelw, XtNfont, font, NULL);
    XtVaSetValues(structlabelw, XtNfont, font, NULL);
    XtVaSetValues(escapelabelw, XtNfont, font, NULL);
    XtVaSetValues(pholdlabelw, XtNfont, font, NULL);
    XtVaSetValues(dumptypelabelw, XtNfont, font, NULL);
  }
  if ((font = CVfont("6x10")) != (XtArgVal) 0) {
    XtVaSetValues(open_win[index].dumptypew, XtNfont, font, NULL);
    XtVaSetValues(open_win[index].stylew, XtNfont, font, NULL);
    XtVaSetValues(open_win[index].fontw, XtNfont, font, NULL);
    XtVaSetValues(open_win[index].structw, XtNfont, font, NULL);
    XtVaSetValues(open_win[index].escapew, XtNfont, font, NULL);
    XtVaSetValues(open_win[index].placeholderw, XtNfont, font, NULL);
    XtVaSetValues(open_win[index].outputw, XtNfont, font, NULL);
  }
  if (open_win[index].textdump->outfile) {
    width = XTextWidth((XFontStruct *) font, open_win[index].textdump->outfile,
		       (strlen(open_win[index].textdump->outfile)));
    XtVaSetValues(open_win[index].outputw, XtNvalue,
		  open_win[index].textdump->outfile,
		  XtNwidth, width, NULL);
  }
  XtAddCallback(open_win[index].dumptypew, XtNcallback, toggle_dumptype,
		(XtPointer) index);
  XtAddCallback(open_win[index].stylew, XtNcallback, toggle_style,
		(XtPointer) index);
  XtAddCallback(open_win[index].fontw, XtNcallback, toggle_font,
		(XtPointer) index);
  XtAddCallback(open_win[index].structw, XtNcallback, toggle_structure,
		(XtPointer) index);
  XtAddCallback(open_win[index].escapew, XtNcallback, toggle_escape,
		(XtPointer) index);
  XtAddCallback(open_win[index].placeholderw, XtNcallback, toggle_phold,
		(XtPointer) index);
  XtAddCallback(doitw, XtNcallback, execute_dump,
		(XtPointer) index);
  XtAddCallback(closew, XtNcallback, close_dump,
		(XtPointer) index);

  XtRealizeWidget(open_win[index].shellw);

  XtVaSetValues(open_win[index].dumptypew, XtNlabel, "LaTeX", NULL);
  XtVaSetValues(open_win[index].structw, XtNlabel, "Deep Structure", NULL);
  XtVaSetValues(open_win[index].escapew, XtNlabel, "Do Escape", NULL);
  XtVaSetValues(open_win[index].placeholderw, XtNlabel, "Include them", NULL);
  XtVaSetValues(open_win[index].fontw, XtNlabel, "medium", NULL);
  XtVaSetValues(open_win[index].stylew, XtNlabel, "article", NULL);

  XtAddCallback(open_win[index].shellw, XtNdestroyCallback, free_stuff,
		(XtPointer) index);
  XtAddCallback(open_win[index].shellw, XtNdestroyCallback, whack_widl,
		(XtPointer) ref);

  XtVaGetValues(open_win[index].shellw, XtNwidth, &width, NULL);
  XtVaSetValues(titlew, XtNwidth, width - 10, NULL);
  XtVaSetValues(open_win[index].statusw, XtNwidth, width - 10, NULL);

  XtPopup(open_win[index].shellw, XtGrabNone);
  return (open_win[index].shellw);
}

/*  */
/*******************************************************************
 * Function: static void num_levels(infonode *inodeptr, int level)
 *
 * This function finds the number of levels in a given reference
 * structure.
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static void num_levels(inodeptr, level)
  infonode *inodeptr;
  int level;
{
  identry *identptr = NULL;
  infonode *infoptr = NULL;

  if ((identptr = inodeptr->subnodes) == (identry *)0) {
    return;
  } else {
    level++;
    if (level > global_level) {
      global_level = level;
    }
    while (identptr != (identry *)0) {
      for (infoptr = identptr->value;
	   infoptr != (infonode *)0; infoptr = infoptr->next) {
	num_levels(infoptr, level);
      }
      identptr = identptr->next;
    }
  }

  if (global_level > 4)
    depth_signal = TRUE;
  return;
}

/*  */
/*******************************************************************
 * Function: static void get_title(FILE *outfile, identry *idptr, int index)
 *
 * This function will find the appropriate title, author, and date
 * for the document and make the title for the LaTeX document.
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static void get_title(outfile, idptr, index)
  FILE *outfile;		/* pointer to LaTeX file */
  identry *idptr;		/* pointer to identry where search starts */
  int index;			/* index into open_win */
{
  char *author = NULL;
  char *title = NULL;
  char *date = NULL;
  identry *saveidptr = idptr;

  (void)check_identry(idptr, &title, &author, &date, index);
  if (!title)
    if (!(title = substitute_title(saveidptr)))
      title = strdup("This should be some title");;
  if (!author)
    author = strdup("This should be your name");
  if (!date)
    date = strdup("This should be the date");

  make_art_title(outfile, title, author, date);

  if (title)
    free((char *)title);
  if (date)
    free((char *)date);
  if (author)
    free((char *)author);
  return;
}

/*  */
/*******************************************************************
 * Function: static int check_identry(FIVE PARAMETERS)
 * Parameters:
 *	identry *idp
 *	char **title
 *	char **author
 *	char **date
 *	int index
 *
 *  go through the identries in a reference structure, calling
 *  check_infonode for each set of infonodes hanging off of the identry.
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static int check_identry(idp, title, author, date, index)
  identry *idp;
  char **title;
  char **author;
  char **date;
  int index;
{
  int rtn = 0;

  while (idp) {
    rtn |= check_infonode(idp->value, title, author, date, index);
    if (rtn)
      return rtn;
    idp = idp->next;
  }
  return rtn;
}

/*  */
/*******************************************************************
 * Function: static int check_infonode(FIVE PARAMETERS)
 * Parameters:
 *	infonode *inp
 *	char **title
 *	char **author
 *	char **date
 *	int index
 *  
 *  at each leaf infonode, check to see if it is appropriate as a
 *  title, author or date.
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static int check_infonode(inp, title, author, date, index)
  infonode *inp;
  char **title;
  char **author;
  char **date;
  int index;
{
  int rtn = 0;
  int found = 0;
  int i;
  int mark = 0;

  while (inp) {
    if ((inp->index == ATTRVALUE) || (inp->index == LINKITEM)) {
      for (i = TITLE; i <= AUTHOR; i++) {
        switch(i) {
        case TITLE:
          if (*title)
            continue;
          break;
        case AUTHOR:
          if (*author)
            continue;
          break;
        case DATE:
          if (*date)
            continue;
          break;
        default:
          break;
        }
	found = compare_strings(inp->field, i, &mark);
	if (found == TRUE) {
	  switch (i) {
	  case TITLE:
	    if (!*title) {
	      *title = strdup(inp->val.attvalue);
	      *title = escape_sp_chars(*title);
	    }
	    break;
	  case AUTHOR:
	    if (!*author) {
	      *author = strdup(inp->val.attvalue);
	      *author = escape_sp_chars(*author);
	    }
	    break;
	  case DATE:
	    if (!*date) {
	      *date = strdup(inp->val.attvalue);
	      *date = escape_sp_chars(*date);
	    }
	    break;
	  default:
	    break;
	  }
	  if (mark)
	    inp->printed = 1;
	}
      }
    }
    rtn |= check_identry(inp->subnodes,title,author,date,index);
    inp = inp->next;
  }
  return rtn;
}


/*  */
/*******************************************************************
 * Function: static char *substitute_title(identry *idptr)
 *
 *  This function will provide a title when none has been found.
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static char *substitute_title(idptr)
  identry *idptr;
{
  char *title = NULL;

  if (!(idptr && idptr->value && idptr->value->val.attsize)) 
    return strdup(idptr->idname);

  if (idptr->value->val.attsize < 30) 
    return strdup(idptr->value->val.attvalue);

  if (!(title = (char *)malloc(30)))
    return NULL;
  (void)strncpy(title, idptr->value->val.attvalue, 25);
  (void)strcat(title, " ...");

  return title;
}

/*  */
/*******************************************************************
 * Function: static int compare_strings(char *string, char *to_find, int *mark)
 *
 * This function does just that.  Only twist is that it checks
 * them in all lower case.  Actually, it looks for a match within
 * the substring of the first argument.
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static int compare_strings(string, to_find, mark)
  char *string;
  int to_find;
  int *mark;			/* whether it should be marked as printed */
{
  char buf[BUF_SIZE];
  int i;
  int index = 0;
  int in_len = 0;
  int iteration = 0;

  static char *authors[] =
  {
    "author",
    "owner"
  };
  static char *dates[] =
  {
    "date",
    "created"
  };

  *mark = 0;
  if (to_find == TITLE)
    in_len = strlen("title");

  for (iteration = 0; iteration < 2; iteration++) {
    index = 0;
    if (to_find == DATE)
      in_len = strlen(dates[iteration]);
    else if (to_find == AUTHOR)
      in_len = strlen(authors[iteration]);

    bzero(buf, BUF_SIZE);
    while (index <= (int) ((strlen(string)) - in_len)) {
      for (i = 0; i < in_len; i++) {	/* string of first in_len length */
	if (islower(string[i + index]))
	  buf[i] = string[i + index];
	else
	  buf[i] = tolower(string[i + index]);
      }
      buf[i] = '\0';
      switch (to_find) {
      case TITLE:
	if (!strcmp("title", buf)) {
	  *mark = 1;
	  return (TRUE);
	}
	break;
      case DATE:
	if (iteration == 0) {
	  if (!strcmp("date", buf)) {
	    *mark = 1;
	    return (TRUE);
	  }
	} else {
	  if (!strcmp("created", buf))
	    return (TRUE);
	}
	break;
      case AUTHOR:
	if (iteration == 0) {
	  if (!strcmp("author", buf)) {
	    *mark = 1;
	    return (TRUE);
	  }
        } else {
	  if (!strcmp("owner", buf))
	    return (TRUE);
	}
	break;
      default:
	break;
      }
      index++;
    }
  }
  return (FALSE);
}

/*  */
/*******************************************************************
 * Function: static void LaTeX_node(int index,infonode *infoptr, FILE *outfile)
 *
 * This function performs a traversal of the reference structure.
 * When a leaf node is encountered, it writes the infonode to the
 * LaTeX file.  It should also maintain a certain sense of
 * hierarchy in the file.
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static void LaTeX_node(index, infoptr, outfile)
  int index;			/* index into open_win */
  infonode *infoptr;		/* pointer to the infonode */
  FILE *outfile;		/* pointer to the LaTeX file */
{
  identry *idptr = NULL;	/* pointer to an identry */
  int where = 0;		/* current position of stackpointer */

  /* get formats whether it's LaTeX or not - for indentation */
  where = add_formats(infoptr->head->tmpl->value);
  (void)add_formats(infoptr);

  if ((idptr = infoptr->subnodes) == (identry *) 0) {
    print_node(index, infoptr, outfile);
  } else {
    while (idptr != (identry *) 0) {
      for (infoptr = idptr->value; infoptr; infoptr = infoptr->next) {
	LaTeX_node(index, infoptr, outfile);
      }
      if (idptr->value->index != ATTRVALUE)
	prevlevel = 0;
      idptr = idptr->next;
    }
  }

  delete_formats(where);
  return;
}

/*  */
/*******************************************************************
 * Function: static void print_node(int index, infonode *info, FILE *outfile)
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static void print_node(index, infoptr, outfile)
  int index;
  infonode *infoptr;
  FILE *outfile;
{
  char *text = NULL;
  char *localfont = NULL;

  if (!infoptr || infoptr->printed)
    return;

  /* print placeholder only if they want it */
  if (infoptr->iflg & NONEXIST_MSK) {
    if(open_win[index].textdump->placeholders == yes) {
      char *text = strdup(infoptr->val.attvalue);
      text = escape_sp_chars(text);
      print_part(index, outfile, text, PARAGRAPH);
      free(text);
    }
    infoptr->printed = 1;
    return;
  }
  
  /* if we're dumping straight text, do it now */
  if (global_type == straight) {
    print_part(index, outfile, infoptr->val.attvalue, PARAGRAPH);
    infoptr->printed = 1;
    return;
  }
  
  if (thislevel != prevlevel) {
    print_part(index, outfile, (char *)0, SQUIGGLES);
  }
  if (stkptr > -1) {
    if ((thislevel != prevlevel) ||
      (prevfont && (strcmp(prevfont, fontstack[stkptr - 1].fontname) != 0))) {
      if (prevfont)
	free((char *)prevfont);
      prevfont = strdup(fontstack[stkptr - 1].fontname);
      localfont = strdup(fontstack[stkptr - 1].fontname);
    }
  }

  if (!(text = attval_plus(infoptr, localfont))) {
    (void)fprintf(stderr, "print_node: attval_plus failed\n");
    return;
  }

  if (open_win[index].textdump->structure == flat) {
    print_part(index, outfile, text, PARAGRAPH);
    infoptr->printed = 1;
  } else {
    thislevel = my_level(infoptr->head);
    if (thislevel != prevlevel) {
      prevlevel = thislevel;
      switch (thislevel) {
      case SECOND_LEVEL:
      case THIRD_LEVEL:
      case FOURTH_LEVEL:
      case FIFTH_LEVEL:
	find_sect_title(index, infoptr, thislevel, outfile);
	if (!infoptr->printed) {
	  print_part(index, outfile, text, PARAGRAPH);
	  infoptr->printed = 1;
	}
	break;
      case FIRST_LEVEL:
      default:
	print_part(index, outfile, text, PARAGRAPH);
	infoptr->printed = 1;
	break;
      }
    } else {
      print_part(index, outfile, text, PARAGRAPH);
      infoptr->printed = 1;
    }
  }

  if (text)
    free((char *)text);
  if (localfont)
    free((char *)localfont);
}

/*  */
/*******************************************************************
 * Function: static char *attval_plus(infonode *infoptr,char *localfont)
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static char *attval_plus(infoptr, localfont)
  infonode *infoptr;
  char *localfont;
{
  char *text = NULL;
  char *newtext = NULL;

  if (!infoptr || !infoptr->val.attvalue)
    return NULL;

  text = strdup(infoptr->val.attvalue);
  text = escape_sp_chars(text);

  if (!(newtext = (char *)malloc((unsigned)strlen(text) +
				 ((localfont) ? strlen(localfont) + 10 : 10)))){
    if (text) 
      free(text);
    return NULL;
  }

  if (localfont) {
    (void)sprintf(newtext, "%s%s", localfont, text);
    lefttoprint += fontstack[stkptr - 1].completion;
  } else {
    (void)strcpy(newtext, text);
  }
  if (text)
    free(text);
  return newtext;
}

/*  */
/*******************************************************************
 * Function: static int my_level(identry *identptr)
 *
 * This function returns the level of the current infonode.  It makes
 * use of the function stepup() from aimeditops.c by Tim Teitenberg.
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static int my_level(idp)
  identry *idp;
{
  int level;

  for (level=0; idp; idp = idp->owner->head, level++)
    if (!idp->owner)
      break;
  return (level);
}

/*  */
/*******************************************************************
 * Function: static char *escape_sp_chars(char *text)
 *
 *  The purpose of this function is to escape special LaTeX characters
 *  before sending the string to be printed to the output file.  At this
 *  point, I will brute force escape them.  This is the solution I
 *  choose simply because there are not terribly many of them and
 *  mapping to functions seems overly complicated.
 *
 *  Arguments:
 *    pointer to the string to be checked.
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static char *escape_sp_chars(text)
  char *text;			/* pointer to string */
{
  char *tmptext = NULL;		/* string that will be sent back eventually */
  char searchchar;		/* character to be checked */
  int index = 0;		/* index into the text string */
  int quotes = LEFT;		/* first quote is a left quote */

  if (!text) {
    (void)fprintf(stderr, "escape_sp_chars: NULL input\n");
    return NULL;
  }

  if (global_escape == no || global_type == straight)
    return text;

  if (!(tmptext = (char *)malloc((unsigned)(strlen(text) * 3))))
    return text;
  bzero(tmptext, strlen(text) * 3);
  for (index = 0; index < (int)strlen(text); index++) {
    searchchar = text[index];
    switch ((int)searchchar) {
    case '"':
      if (quotes == LEFT) {
	(void)strcat(tmptext, "``");
	quotes = RIGHT;
      } else {
	(void)strcat(tmptext, "''");
	quotes = LEFT;
      }
      break;
    case '\\':
      (void)strcat(tmptext, "$\\backslash{}$");
      break;
    case '~':
      (void)strcat(tmptext, "\\~{}");
      break;
    case '^':
      (void)strcat(tmptext, "\\^{}");
      break;
    case '<':
      (void)strcat(tmptext, "$<$");
      break;
    case '>':
      (void)strcat(tmptext, "$>$");
      break;
    case '|':
      (void)strcat(tmptext, "$\\mid$");
      break;
    case '{':
      (void)strcat(tmptext, "\\{");
      break;
    case '}':
      (void)strcat(tmptext, "\\}");
      break;
    case '$':
    case '%':
    case '&':
    case '#':
    case '_':			/* perform the same on all */
      (void)strcat(tmptext, "\\");
    default:
      (void)strcat(tmptext, &searchchar);
      break;
    }
  }
  free((char *)text);
  return tmptext;
}

/*  */
/*******************************************************************
 * Function: static void find_sect_title(FOUR PARAMETERS)
 * Parameters:
 *	int index		into open_win
 *	infonode *infoptr
 *	int check_level
 *	FILE *outfile
 *
 * This function will find the appropriate title to give to the
 * heading of the current level section (section, subsection, etc.).
 * If there is an ATTRVALUE at this level, it will check to see if
 * the infonode->field has the string "title" in any part of it.
 * I will essentially to a "window idiom" search and compare of the
 * string.  If I find something at this level, it will be used.
 * Otherwise, I will give the section the title "Temporary Section
 * Title" and let the user change it in the file as s/he sees fit.
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static void find_sect_title(index, infoptr, check_level, outfile)
  int index;			/* into open_win */
  infonode *infoptr;		/* ptr to infonode where you begin */
  int check_level;		/* current level. only check this level */
  FILE *outfile;		/* output file */
{
  infonode *tmpinfoptr = NULL;
  identry *tmpidptr = NULL;
  int found = FALSE;
  char *title = NULL;
  int mark = 0;			/* passed to compare_strings(), used
				 * to signal whether an infonode should
				 * be marked as printed or not
				 */

/*
 * look at the first infonode at this level of the reference
 * structure.  If it has the string "title" in the "field" field of the
 * infonode, then you have found your section title.
 */
  tmpidptr = infoptr->head;
  if (tmpidptr) {
    tmpinfoptr = tmpidptr->value;
    if (tmpinfoptr->index == ATTRVALUE)
      found = compare_strings(tmpinfoptr->field, TITLE, &mark);
  }
  if (found) {
    title = strdup(tmpinfoptr->val.attvalue);
    title = escape_sp_chars(title);
    lefttoprint = 0;
  }
  switch (check_level) {
  case SECOND_LEVEL:
    if (found && depth_signal)
      print_part(index, outfile, title, PART);
    else if (found)
      print_part(index, outfile, title, SECTION);
    else
      print_part(index, outfile, "Temporary Section Title", SECTION);
    break;
  case THIRD_LEVEL:
    if (found && depth_signal)
      print_part(index, outfile, title, SECTION);
    else if (found == TRUE)
      print_part(index, outfile, title, SUBSECTION);
    else
      print_part(index, outfile, "Temporary Subsection Title", SUBSECTION);
    break;
  case FOURTH_LEVEL:
    if (found && depth_signal)
      print_part(index, outfile, title, SUBSECTION);
    else if (found == TRUE)
      print_part(index, outfile, title, SUBSUBSECTION);
    else
      print_part(index, outfile, "Temporary Subsubection Title", SUBSUBSECTION);
    break;
  case FIFTH_LEVEL:
    if (found && depth_signal)
      print_part(index, outfile, title, SUBSUBSECTION);
    break;
  case FIRST_LEVEL:
  default:
    (void)fprintf(stderr, "find_sect_title:  no such case\n");
    break;
  }
  if (mark)
    tmpinfoptr->printed = 1;

  if (title)
    free((char *)title);
  return;
}

/*  */
/*******************************************************************
 * Function: static void clear_printed(infonode *inodeptr)
 *
 * This function does a traversal of the reference structure to clear
 * all of the "printed" fields in the infonodes.  This is done so that
 * the next time the file is LaTeX'ed, the information will be valid.
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static void clear_printed(inodeptr)
  infonode *inodeptr;
{
  identry *identptr = NULL;
  infonode *infoptr = NULL;

  if ((identptr = inodeptr->subnodes) == (identry *)0) {
    inodeptr->printed = FALSE;
    return;
  } else {
    while (identptr != (identry *)0) {
      for (infoptr = identptr->value;
	   infoptr != (infonode *)0;
	   infoptr = infoptr->next) {
	clear_printed(infoptr);
      }
      identptr = identptr->next;
    }
  }
  return;
}

/*  */
/**********************************************************************
 * Function: static int add_formats(infonode *infoptr)
 *
 * This function adds fonts from generic and specific fonting
 * information to a stack of fonts.
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static int add_formats(infoptr)
  infonode *infoptr;
{
  formentry *tmp = NULL;
  filefontinfo *f = NULL;
  int level = 0;

  if (stkptr == -1)
    level = 0;
  else
    level = stkptr;

  for (tmp = infoptr->formats; tmp; tmp = tmp->next) {
    if (strcmp(tmp->formid, "font") == 0) {
      f = get_format(tmp->formval);
      if (f) {
	if (stkptr == -1)
	  stkptr = 0;
	if (fontstack[stkptr].fontname)
	  free((char *)fontstack[stkptr].fontname);
	fontstack[stkptr].fontname = strdup(f->LaTeXname);
	fontstack[stkptr].completion = f->squigglies;
	stkptr++;
      }
    }
    if (strcmp(tmp->formid, "leftMargin") == 0) {
      int tmpint = atoi(tmp->formval);
      indent = (int)(tmpint / 10);
    }
  }
  return (level);
}

/*  */
/**********************************************************************
 * Function: static void delete_formats(int where)
 *
 * This function puts the stack of fonts back to the state it was
 * in at the invocation of the function.
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static void delete_formats(where)
  int where;
{
  int i = 0;

  for (i = where; i < stkptr; i++) {
    if (fontstack[i].fontname)
      free((char *)fontstack[i].fontname);
    fontstack[i].fontname = NULL;
    fontstack[i].completion = 0;
  }
  if (where == 0)
    stkptr = -1;
  else
    stkptr = where;
}

/*  */
/**********************************************************************
 * Function: static int load_file_fonts()
 *
 * load all of the font translations from file.  this is done
 * once per xlincks session the first time that the text dumping
 * operation is carried out.
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static int load_file_fonts()
{
  FILE *fontfile = NULL;
  filefontinfo *tmp = NULL;
  int i = 0, j = 0;
  char f1[BUF_SIZE];
  char f2[BUF_SIZE];
  static int done = 0;
  char filename[MAXPATHLEN];

  (void)sprintf(filename, "%s/%s", DBDIR, FONT_TRANSLATION_FILE);

  if (!(fontfile = fopen(filename, "r"))) {
    (void)fprintf(stderr, 
		  "Cannot open %s\nfor X to LaTeX translations.\n", filename);
    (void)fprintf(stderr, 
		  "Contact your systems administrator to get it installed.\n");
    return (0);
  }

  if (done)
    return 1;

  done = 1;
  bzero(f1, sizeof(f1));
  bzero(f2, sizeof(f2));

  while ((i = fscanf(fontfile, "%s %s %d", f1, f2, &j)) != EOF) {
    if (i < 3)
      break;
    tmp = new_filefontinfo();
    tmp->Xfontname = strdup(f1);
    tmp->LaTeXname = strdup(f2);
    tmp->squigglies = j;
    if (!fontinfo)
      fontinfo = tmp;
    else {
      tmp->next = fontinfo;
      fontinfo = tmp;
    }
    bzero(f1, sizeof(f1));
    bzero(f2, sizeof(f2));
  }
  if ((fclose(fontfile)) != 0) {
    (void)fprintf(stderr, "fclose failed in load_file_fonts\n");
  }
  return 1;
}

/*  */
/**********************************************************************
 * Function: static filefontinfo *new_filefontinfo()
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static filefontinfo *new_filefontinfo()
{
  filefontinfo *tmp = NULL;

  tmp = (filefontinfo *) malloc(sizeof(filefontinfo));
  tmp->next = NULL;
  tmp->Xfontname = NULL;
  tmp->LaTeXname = NULL;
  tmp->squigglies = 0;
  return tmp;
}

/*  */
/**********************************************************************
 * Function: static filefontinfo *get_format(char *lookfor)
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static filefontinfo *get_format(lookfor)
  char *lookfor;
{
  filefontinfo *f = NULL;

  for (f = fontinfo; f; f = f->next) {
    if (!(strcmp(lookfor, f->Xfontname)))
      return f;
  }
  return NULL;
}


/*  */
/**********************************************************************
 * Function: static static _textdump *initialize()
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static _textdump *initialize()
{
  static int done = 0;
  _textdump *textdump = NULL;

  if (!done) {
    int i;
    for (i = 0; i < MAX_WINS; i++) {
      open_win[i].ref = NULL;
      open_win[i].textdump = NULL;
      open_win[i].dumptypew = NULL;
      open_win[i].stylew = open_win[i].fontw = NULL;
      open_win[i].escapew = open_win[i].structw = open_win[i].shellw = NULL;
      open_win[i].placeholderw=open_win[i].statusw=open_win[i].outputw = NULL;
    }
    done = 1;
  }
  textdump = (_textdump *) malloc(sizeof(_textdump));
  textdump->outfile = get_file_name();
  textdump->docstyle = article;
  textdump->structure = deep;
  textdump->fontsize = medium;
  textdump->placeholders = yes;
  global_escape = doit;
  global_type = LaTeX;

  if (!load_file_fonts())
    (void)fprintf(stderr, "load_file_fonts: cannot open fontfile.\n");

  return textdump;
}

/*  */
/**********************************************************************
 * Function: static int is_open(reference_structure *ref, _textdump *textdump)
 *
 * go through the open_win structure and see if we have a text dump
 * window open for the reference structure.  if so, don't open another,
 * and if not find an empty slot and slam in ref and textdump in the
 * right places.
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static int is_open(ref, textdump)
  reference_structure *ref;
  _textdump *textdump;
{
  int i;

  for (i = 0; i < MAX_WINS; i++)
    if (open_win[i].ref == ref) {
      XtVaSetValues(open_win[i].statusw, XtNlabel,
		  "Sorry.  Only one Dump Window per composite object.", NULL);
      XBell(XtDisplay(open_win[i].statusw), 50);
      if (textdump->outfile) {
	free((char *)textdump->outfile);
	textdump->outfile = NULL;
      }
      if (textdump) {
	free((char *)textdump);
	textdump = NULL;
      }
      return -1;
    }
  for (i = 0; i < MAX_WINS; i++) {
    if (!open_win[i].ref) {
      open_win[i].ref = ref;
      open_win[i].textdump = textdump;
      return i;
    }
  }
  popup_message(POPUP_MESS_INFO, "%s %d %s", "You can only have",
		MAX_WINS, "Text Dump Windows open at the same time.");
  if (textdump->outfile) {
    free((char *)textdump->outfile);
    textdump->outfile = NULL;
  }
  if (textdump) {
    free((char *)textdump);
    textdump = NULL;
  }
  return -1;
}

/*  */
/**********************************************************************
 * Function: static int set_outfile(int index)
 *
 * get the string value from the dialog widget, ensure it's at least
 * more than just a null char at the beginning.  that string is
 * managed by X, so it cannot be free'ed.  i keep my own copy.
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static int set_outfile(index)
  int index;			/* into open_win */
{
  char *filename = NULL;

  filename = XawDialogGetValueString(open_win[index].outputw);
  if ((!filename) || (filename[0] == '\0'))
    return 0;

  if (open_win[index].textdump->outfile)
    free((char *)open_win[index].textdump->outfile);
  open_win[index].textdump->outfile = strdup(filename);
  return 1;
}

/*  */
/**********************************************************************
 * Function: static char *get_file_name()
 *
 * choose a suitable file for the default output
 *
 * Modifications:
 *      <list mods with name and date>
 */
static char *get_file_name()
{
  struct stat buf;		/* stat information returned here */
  char *home = NULL;
  char *base = NULL;
  char *try = NULL;
  int done = 0;
  int count = 0;
  extern char *getenv();

  if ((home = (char *)getenv("HOME")) != NULL) {
    base = (char *)malloc((unsigned)strlen(home) + 20);
    (void)sprintf(base, "%s%s", home, "/");
  } else {
    base = (char *)malloc((unsigned)25);	/* /tmp/ + some */
    (void)sprintf(base, "%s", "/tmp/");
  }

  try = (char *)malloc((unsigned)strlen(base) + 15);
  (void)sprintf(try, "%s%s%s", base, "tmpfile", ".tex");
  while (!done) {
    if (stat(try, &buf) != -1) {
#ifdef SPRINTF_INT
      if (sprintf(try, "%s%s%d%s", base, "tmpfile", ++count, ".tex") == EOF)
#else
      if (sprintf(try, "%s%s%d%s", base, "tmpfile", ++count, ".tex") == NULL)
#endif
	(void)fprintf(stderr, "get_file_name: sprintf failed.\n");
    } else {
      free((char *)base);
      return try;
    }
  }

  free((char *)base);
  return try;
}

/*  */
/**********************************************************************
 * Function: static void get_header(int index, FILE *outfile)
 *
 * This function will include the header for the LaTeX file.
 * This header is partly included in the global formating attributes
 * in formats for each document.  The type size and
 * document style are taken from that structure.
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void get_header(index, outfile)
  int index;
  FILE *outfile;
{
  static char *dockinds[] = {
    "article",
    "letter"
  };

  char *doc = (char *)dockinds[(int)open_win[index].textdump->docstyle];
  int font = (int)open_win[index].textdump->fontsize;

  if (open_win[index].textdump->docstyle == article)
    (void)fprintf(outfile, "%s%d%s%s%s\n", "\\documentstyle[",
		  font, "pt,a4,par]{", doc, "}");
  else
    (void)fprintf(outfile, "%s%d%s%s%s\n",
		  "\\documentstyle[", font, "pt]{", doc, "}");

  (void)fprintf(outfile, "%s\n%s\n%s\n",
		"\\marginparwidth 0pt",
		"\\oddsidemargin  35pt",
		"\\evensidemargin  35pt");
  (void)fprintf(outfile, "%s\n%s\n",
		"\\marginparsep 0pt",
		"\\topmargin   0pt");
  (void)fprintf(outfile, "%s\n%s\n%s\n%s\n",
		"\\textwidth=422pt",
		"\\textheight=25cm",
		"\\parindent=0cm  % if you want indention, change this",
		"\\headsep=1cm");
  (void)fprintf(outfile, "%s\n\n",
		"\\begin{document}");
#ifdef DEBUG
  if (open_win[index].textdump->docstyle == article)
    (void)printf("%s%d%s%s%s\n", "\\documentstyle[", font, "pt,a4,par]{", doc, "}");
  else
    (void)printf("%s%d%s%s%s\n", "\\documentstyle[", font, "pt]{", doc, "}");
  (void)printf("%s\n%s\n%s\n",
	       "\\marginparwidth 0pt",
	       "\\oddsidemargin  35pt",
	       "\\evensidemargin  35pt");
  (void)printf("%s\n%s\n",
	       "\\marginparsep 0pt",
	       "\\topmargin   0pt");
  (void)printf("%s\n%s\n%s\n%s\n",
	       "\\textwidth=422pt",
	       "\\textheight=25cm",
	       "\\parindent=0cm  % if you want indention, change this",
	       "\\headsep=1cm");
  (void)printf("%s\n\n",
	       "\\begin{document}");
#endif
}

/*  */
/**********************************************************************
 * Function: static void make_art_title(FOUR PARAMETERS)
 * Parameters:
 *	FILE *outfile
 *	char *title
 *	char *author
 *	char *date
 *
 *  This function inserts text at the paragraph level.  It will grab
 *  the appropriate text from the appropriate place and put it into the
 *  file where it should be.
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void make_art_title(outfile, title, author, date)
  FILE *outfile;		/* the output LaTeX file */
  char *title;			/* the title of the document */
  char *author;			/* the author of the document */
  char *date;			/* the date of the document */
{
  (void)fprintf(outfile, "%s%s%s\n%s%s%s\n%s%s%s\n\n%s",
		"\\title{", title, "}",
		"\\author{", author, "}",
		"\\date{", date, "}",
		"\\maketitle");
#ifdef DEBUG
  (void)printf("%s%s%s\n%s%s%s\n%s%s%s\n\n%s\n",
	       "\\title{", title, "}",
	       "\\author{", author, "}",
	       "\\date{", date, "}",
	       "\\maketitle");
#endif

}

/*  */
/**********************************************************************
 * Function: static void print_indent(FILE *outfile, int debug)
 *
 * Print out some spaces (there must have been a leftMargin formatting
 * directive).
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void print_indent(outfile, debug)
  FILE *outfile;
  int debug;
{
  int i;

  for (i = 0; i < indent; i++) {
    if (debug)
      (void)printf("  ");
    else
      (void)fprintf(outfile, "  ");
  }
}

/*  */
/**********************************************************************
 * Function: static void print_part(FOUR PARAMETERS)
 * Parameters:
 *	int index		- index into open_win
 *	FILE *outfile
 *	char *part		-string to print
 *	int which		- which part
 *
 *  This function prints some part of the LaTeX to the output
 *  file.
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void print_part(index, outfile, part, which)
  int index;			/* into open_win */
  FILE *outfile;		/* the output LaTeX file */
  char *part;			/* the title of the part */
  int which;
{
  int j;

  (void)fprintf(outfile, "\n\n");
#ifdef DEBUG
  (void)printf("\n\n");
#endif

  switch (which) {
  case PART:
    (void)fprintf(outfile, "%s", "\\part{");
#ifdef DEBUG
    (void)printf("%s", "\\part{");
#endif
    break;
  case SECTION:
    (void)fprintf(outfile, "%s", "\\section{");
#ifdef DEBUG
    (void)printf("%s", "\\section{");
#endif
    break;
  case SUBSECTION:
    (void)fprintf(outfile, "%s", "\\subsection{");
#ifdef DEBUG
    (void)printf("%s", "\\subsection{");
#endif
    break;
  case SUBSUBSECTION:
    (void)fprintf(outfile, "%s", "\\subsubsection{");
#ifdef DEBUG
    (void)printf("%s", "\\subsubsection{");
#endif
    break;
  default:
    break;
  }

  switch (which) {
  case PART:
  case SECTION:
  case SUBSECTION:
  case SUBSUBSECTION:
  case PARAGRAPH:
    print_indent(outfile, 0);
    (void)fprintf(outfile, "%s%s", part, (which == PARAGRAPH) ? " " : "}");
#ifdef DEBUG
    print_indent(outfile, 1);
    (void)printf("%s%s", part, (which == PARAGRAPH) ? " " : "}");
#endif
    break;
  case TRAILER:
    if (global_type == LaTeX) {
      (void)fprintf(outfile, "\\end{document}\n");
#ifdef DEBUG
      (void)printf("\\end{document}\n");
#endif
    } else {
      (void)fprintf(outfile, "\n");
#ifdef DEBUG
      (void)printf("\n");
#endif
    }
    break;
  case LETTERHEAD:
    (void)fprintf(outfile, "%s\n\n%s\n\n%s\n\n%s",
		  "\\signature{Your name goes here}",
		  "\\address{Your address\\\\ should go here}",
		  "\\begin{letter}{The recipients address\\\\ should go here}",
		  "\\opening{Dear Somebody}   % the salutation");
#ifdef DEBUG
    (void)printf("%s\n\n%s\n\n%s\n\n%s\n\n",
		 "\\signature{Your name goes here}",
		 "\\address{Your address \\\\ should go here}",
		 "\\begin{letter}{The recipients address \\\\ should go here}",
		 "\\opening{Dear Somebody}  % the salutation");
#endif
    break;
  case LETTERTAIL:
    (void)fprintf(outfile, "%s\n\n%s\n",
		  "\\closing{Sincerely}  % You can change this here",
		  "\\end{letter}");
#ifdef DEBUG
    (void)printf("%s\n\n%s\n",
		 "\\closing{Sincerely}  % You can change this here",
		 "\\end{letter}");
#endif
    break;
  case SQUIGGLES:
    for (j = 0; j < lefttoprint; j++) {
      (void)fprintf(outfile, "%s", "}");
#ifdef DEBUG
      (void)printf("%s", "}");
#endif
    }
    lefttoprint = 0;
    break;
  default:
    (void)fprintf(stderr, "print_part: invalid case\n");
    break;
  }
  indent = 0;	/* no. of times to indent */
}

/*  */
/* below are callbacks of various sorts for the widgets.  most are
 * for toggling in the command widgets, but there are also a couple
 * of destroy callbacks
 */
/* ARGSUSED */
/*******************************************************************
 * Function: static void execute_dump(CALLBACK PROTOCOL)
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static void execute_dump(widget, client_data, call_data)
  Widget widget;
  XtPointer client_data, call_data;
{
  int index = (int)client_data;
  FILE *outfile;		/* pointer to the LaTeX file */
  int level = 0;		/* dummy variable passed to num_levels */

  if (!set_outfile(index)) {
    (void)fprintf(stderr, "Cannot have NULL output file.\n");
    XtVaSetValues(open_win[index].statusw, XtNlabel,
		  "Error: Cannot have NULL output file", NULL);
    XBell(XtDisplay(widget), 50);
    return;
  }
  if ((outfile = fopen(open_win[index].textdump->outfile, "w")) == (FILE *) 0) {
    (void)fprintf(stderr, "Unable to open output file.\n");
    XtVaSetValues(open_win[index].statusw, XtNlabel,
		  "Error: Unable to open output file", NULL);
    XBell(XtDisplay(widget), 50);
    return;
  }
  num_levels((open_win[index].ref)->target->value, level);

  if (global_type == LaTeX) {
    get_header(index, outfile);
    if (open_win[index].textdump->docstyle == article)
      get_title(outfile, (open_win[index].ref)->target, index);
    else
      print_part(index, outfile, (char *)0, LETTERHEAD);
  }
  LaTeX_node(index, (open_win[index].ref)->target->value, outfile);

  if (global_type == LaTeX)
    print_part(index, outfile, (char *)0, SQUIGGLES);
  if ((global_type == LaTeX) && 
      (open_win[index].textdump->docstyle == letter))
    print_part(index, outfile, (char *)0, LETTERTAIL);
  print_part(index, outfile, (char *)0, TRAILER);

  if ((fclose(outfile)) != 0)
    (void)fprintf(stderr, "fclose failed in execute_dump.\n");

  /* clear "printed" in infonodes */
  clear_printed((open_win[index].ref)->target->value);

  XtVaSetValues(open_win[index].statusw,
		XtNlabel, "Status: Output sent to file successfully.", NULL);
  XBell(XtDisplay(widget), 50);
}

/*  */
/* ARGSUSED */
/**********************************************************************
 * Function: static void toggle_dumptype(CALLBACK PROTOCOL)
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static void toggle_dumptype(widget, client_data, call_data)
  Widget widget;
  XtPointer client_data, call_data;
{
  int index = (int)client_data;
  Widget w = (Widget) open_win[index].dumptypew;

  switch (global_type) {
  case straight:
    global_type = LaTeX;
    XtVaSetValues(w, XtNlabel, "LaTeX", NULL);
    break;
  case LaTeX:
    global_type = straight;
    XtVaSetValues(w, XtNlabel, "Straight Text", NULL);
    break;
  default:
    (void)printf("toggle_dumptype: unknown dump type\n");
    break;
  }
}

/*  */
/* ARGSUSED */
/**********************************************************************
 * Function: static void toggle_style(CALLBACK PROTOCOL)
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static void toggle_style(widget, client_data, call_data)
  Widget widget;
  XtPointer client_data, call_data;
{
  int index = (int)client_data;
  Widget w = (Widget) open_win[index].stylew;

  switch (open_win[index].textdump->docstyle) {
  case article:
    open_win[index].textdump->docstyle = letter;
    XtVaSetValues(w, XtNlabel, "letter", NULL);
    break;
  case letter:
    open_win[index].textdump->docstyle = article;
    XtVaSetValues(w, XtNlabel, "article", NULL);
    break;
  default:
    (void)printf("toggle_style: unknown document style\n");
    break;
  }
}

/*  */
/* ARGSUSED */
/**********************************************************************
 * Function: static void toggle_font(CALLBACK PROTOCOL)
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static void toggle_font(widget, client_data, call_data)
  Widget widget;
  XtPointer client_data, call_data;
{
  int index = (int)client_data;
  Widget w = (Widget) open_win[index].fontw;

  switch (open_win[index].textdump->fontsize) {
  case small:
    open_win[index].textdump->fontsize = medium;
    XtVaSetValues(w, XtNlabel, "medium", NULL);
    break;
  case medium:
    open_win[index].textdump->fontsize = large;
    XtVaSetValues(w, XtNlabel, "large", NULL);
    break;
  case large:
    open_win[index].textdump->fontsize = small;
    XtVaSetValues(w, XtNlabel, "small", NULL);
    break;
  default:
    (void)printf("toggle_font: unknown font size\n");
    break;
  }
}

/*  */
/* ARGSUSED */
/**********************************************************************
 * Function: static void toggle_structure(CALLBACK PROTOCOL)
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static void toggle_structure(widget, client_data, call_data)
  Widget widget;
  XtPointer client_data, call_data;
{
  int index = (int)client_data;
  Widget w = (Widget) open_win[index].structw;

  switch (open_win[index].textdump->structure) {
  case flat:
    open_win[index].textdump->structure = deep;
    XtVaSetValues(w, XtNlabel, "Deep Structure", NULL);
    break;
  case deep:
    open_win[index].textdump->structure = flat;
    XtVaSetValues(w, XtNlabel, "Flat Structure", NULL);
    break;
  default:
    (void)printf("toggle_structure: unknown structure type\n");
    break;
  }
}

/*  */
/* ARGSUSED */
/**********************************************************************
 * Function: static void toggle_phold(CALLBACK PROTOCOL)
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static void toggle_phold(widget, client_data, call_data)
  Widget widget;
  XtPointer client_data, call_data;
{
  int index = (int)client_data;
  Widget w = (Widget) open_win[index].placeholderw;

  switch (open_win[index].textdump->placeholders) {
  case yes:
    open_win[index].textdump->placeholders = no;
    XtVaSetValues(w, XtNlabel, "No placeholders", NULL);
    break;
  case no:
    open_win[index].textdump->placeholders = yes;
    XtVaSetValues(w, XtNlabel, "Include them", NULL);
    break;
  default:
    (void)printf("toggle_phold: unknown escape type\n");
    break;
  }
}

/*  */
/* ARGSUSED */
/**********************************************************************
 * Function: static void toggle_escape(CALLBACK PROTOCOL)
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static void toggle_escape(widget, client_data, call_data)
  Widget widget;
  XtPointer client_data, call_data;
{
  int index = (int)client_data;
  Widget w = (Widget) open_win[index].escapew;

  switch (global_escape) {
  case dont:
    global_escape = doit;
    XtVaSetValues(w, XtNlabel, "Do Escape", NULL);
    break;
  case doit:
    global_escape = dont;
    XtVaSetValues(w, XtNlabel, "Don't escape", NULL);
    break;
  default:
    (void)printf("toggle_escape: unknown escape type\n");
    break;
  }
}

/*  */
/* ARGSUSED */
/**********************************************************************
 * Function: static void free_stuff(CALLBACK PROTOCOL)
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static void free_stuff(widget, client_data, call_data)
  Widget widget;
  XtPointer client_data, call_data;
{
  int index = (int)client_data;

  if (open_win[index].textdump->outfile)
    free((char *)open_win[index].textdump->outfile);
  if (open_win[index].textdump)
    free((char *)open_win[index].textdump);
  if (prevfont)
    free((char *)prevfont);
  prevfont = NULL;
  open_win[index].ref = NULL;
  open_win[index].textdump = NULL;
  open_win[index].stylew = open_win[index].dumptypew = NULL;
  open_win[index].fontw = open_win[index].structw = NULL;
  open_win[index].escapew = open_win[index].placeholderw = NULL;
  open_win[index].shellw = open_win[index].statusw = NULL;
  open_win[index].outputw = NULL;
}

/*  */
/* ARGSUSED */
/**********************************************************************
 * Function: static void close_dump(CALLBACK PROTOCOL)
 *
 *  Modifications:
 *	<list mods with name and date>
 */
static void close_dump(widget, client_data, call_data)
  Widget widget;
  XtPointer client_data, call_data;
{
  int index = (int)client_data;
  XtDestroyWidget(open_win[index].shellw);
}

